@@ -1,9 +1,9 @@
 package Net::DNS::RR;
 
 #
-# $Id: RR.pm 1530 2017-01-27 10:40:37Z willem $
+# $Id: RR.pm 1597 2017-09-22 08:04:02Z willem $
 #
-our $VERSION = (qw$LastChangedRevision: 1530 $)[1];
+our $VERSION = (qw$LastChangedRevision: 1597 $)[1];
 
 
 =head1 NAME
@@ -110,7 +110,7 @@ sub _new_string {
 	my $t2 = uc $token[1] if $#token;
 
 	my ( $ttl, $class );
-	unless ( defined $t2 ) {				# <owner> <type>
+	if ( not defined $t2 ) {				# <owner> <type>
 		@token = ('ANY') if $classbyname{$t1};		# <owner> <class>
 	} elsif ( $classbyname{$t1} || $t1 =~ /^CLASS\d/ ) {
 		$class = shift @token;				# <owner> <class> [<ttl>] <type>
@@ -132,11 +132,10 @@ sub _new_string {
 
 	if ( $#token && $token[0] =~ /^[\\]?#$/ ) {
 		shift @token;					# RFC3597 hexadecimal format
-		my $count = shift(@token) || 0;
-		my $rdata = pack 'H*', join '', @token;
-		my $rdlen = $self->{rdlength} = length $rdata;
-		croak 'length and hexadecimal data inconsistent' unless $rdlen == $count;
-		$self->_decode_rdata( \$rdata, 0 ) if $rdlen;	# unpack RDATA
+		my $rdlen = shift(@token) || 0;
+		my $rdata = pack 'H*', join( '', @token );
+		croak 'length and hexadecimal data inconsistent' unless $rdlen == length $rdata;
+		$self->rdata($rdata) if $rdlen;			# unpack RDATA
 		return $self;
 	}
 
@@ -173,28 +172,30 @@ sections required for certain dynamic update operations.
 
 =cut
 
-sub _new_hash {
-	my ( $base, %argument ) = @_;
+my @core = qw(owner name type class ttl rdlength);
 
-	my %attribute = ( owner => '.', type => 'ANY' );
-	while ( my ( $key, $value ) = each %argument ) {
-		$attribute{lc $key} = $value;
+sub _new_hash {
+	my $base = shift;
+
+	my %attribute = ( owner => '.', type => 'NULL' );
+	while ( my $key = shift ) {
+		$attribute{lc $key} = shift;
 	}
 
-	my ( $owner, $type, $class, $ttl ) = @attribute{qw(owner type class ttl)};
-	$owner = $attribute{name} if exists $attribute{name};	# synonym for owner
+	my ( $owner, $name, $type, $class, $ttl ) = @attribute{@core};
+	delete @attribute{@core};				# leaving RDATA only
 
-	delete @attribute{qw(owner name class type ttl rdlength)};
-	my $populated = scalar %attribute;			# RDATA specified
-
-	my $self = $base->_subclass( $type, $populated );	# RR with defaults (if appropriate)
-	$self->owner($owner);
+	my $self = $base->_subclass( $type, scalar %attribute );
+	$self->owner( $name ? $name : $owner );
 	$self->class($class) if defined $class;			# specify CLASS
 	$self->ttl($ttl)     if defined $ttl;			# specify TTL
 
-	while ( my ( $attribute, $value ) = each %attribute ) {
-		$self->$attribute( ref($value) eq 'ARRAY' ? @$value : $value );
-	}
+	eval {
+		while ( my ( $attribute, $value ) = each %attribute ) {
+			$self->$attribute( ref($value) eq 'ARRAY' ? @$value : $value );
+		}
+	};
+	die ref($self) eq __PACKAGE__ ? "type $type not implemented" : () if $@;
 
 	return $self;
 }
@@ -271,7 +272,7 @@ sub encode {
 	my $type  = $self->{type};
 	my $class = $self->{class} || 1;
 	my $index = $offset + length($owner) + RRFIXEDSZ;
-	my $rdata = eval { $self->_encode_rdata( $index, @opaque ); } || '';
+	my $rdata = eval { $self->_empty ? '' : $self->_encode_rdata( $index, @opaque ); } || '';
 	return pack 'a* n2 N n a*', $owner, $type, $class, $self->ttl, length $rdata, $rdata;
 }
 
@@ -296,7 +297,7 @@ sub canonical {
 	my $type  = $self->{type};
 	my $class = $self->{class} || 1;
 	my $index = RRFIXEDSZ + length $owner;
-	my $rdata = eval { $self->_encode_rdata($index); } || '';
+	my $rdata = eval { $self->_empty ? '' : $self->_encode_rdata($index); } || '';
 	pack 'a* n2 N n a*', $owner, $type, $class, $self->ttl, length $rdata, $rdata;
 }
 
@@ -333,11 +334,12 @@ sub string {
 	my @ttl	 = grep defined, $self->{ttl};
 	my @core = ( $name, @ttl, $self->class, $self->type );
 
-	my @rdata = eval { $self->_format_rdata; };
+	my $empty = $self->_empty;
+	my @rdata = eval { $empty ? () : $self->_format_rdata; };
 	carp $@ if $@;
 
 	my $tab = length($name) < 72 ? "\t" : ' ';
-	return join $tab, @core, '; no data' unless scalar @rdata;
+	$self->_annotation('no data') if $empty;
 
 	my @line = _wrap( join( $tab, @core, '(' ), @rdata, ')' );
 
@@ -377,14 +379,11 @@ sub token {
 	my @ttl = grep defined, $self->{ttl};
 	my @core = ( $self->{owner}->string, @ttl, $self->class, $self->type );
 
+	my @rdata = eval { $self->_empty ? () : $self->_format_rdata; };
+
 	# parse into quoted strings, contiguous non-whitespace and (discarded) comments
-	local $_ = join ' ', eval { $self->_format_rdata; };
-	s/\\\\/\\092/g;						# disguise escaped escape
-	s/\\"/\\034/g;						# disguise escaped quote
-	s/\\\(/\\040/g;						# disguise escaped bracket
-	s/\\\)/\\041/g;						# disguise escaped bracket
-	s/\\;/\\059/g;						# disguise escaped semicolon
-	my @token = @core, grep defined && length, split /$PARSE_REGEX/o;
+	my @parse = map { s/\\\\/\\092/g; s/\\"/\\034/g; split /$PARSE_REGEX/o; } @rdata;
+	my @token = ( @core, grep defined && length, @parse );
 }
 
 
@@ -454,8 +453,8 @@ Resource record class.
 
 sub class {
 	my $self = shift;
-	$self->{class} = classbyname(shift) if scalar @_;
-	classbyval( $self->{class} || 1 ) if defined wantarray;
+	return $self->{class} = classbyname(shift) if scalar @_;
+	defined $self->{class} ? classbyval( $self->{class} ) : 'IN';
 }
 
 
@@ -472,7 +471,6 @@ Resource record time to live in seconds.
 # published API.  These are required for parsing BIND zone files but
 # should not be used in other contexts.
 my %unit = ( W => 604800, D => 86400, H => 3600, M => 60, S => 1 );
-%unit = ( %unit, map /\D/ ? lc($_) : $_, %unit );
 
 sub ttl {
 	my ( $self, $time ) = @_;
@@ -482,7 +480,7 @@ sub ttl {
 	my $ttl = 0;
 	my %time = reverse split /(\D)\D*/, $time . 'S';
 	while ( my ( $u, $t ) = each %time ) {
-		my $scale = $unit{$u} || die qq(bad time: $t$u);
+		my $scale = $unit{uc $u} || die qq(bad time: $t$u);
 		$ttl += $t * $scale;
 	}
 	$self->{ttl} = $ttl;
@@ -503,21 +501,19 @@ sub _decode_rdata {			## decode rdata from wire-format octet string
 
 sub _encode_rdata {			## encode rdata as wire-format octet string
 	my $rdata = shift->{rdata};
-	return defined $rdata ? $rdata : '';
 }
 
 
 sub _format_rdata {			## format rdata portion of RR string
-	my $self = shift;
-	my $data = $self->rdata;
-	my $size = length($data) || return '';			# RFC3597 unknown RR format
+	my $data = shift->rdata;
+	my $size = length($data);				# RFC3597 unknown RR format
 	my @data = ( '\\#', $size, split /(\S{32})/, unpack 'H*', $data );
 }
 
 
 sub _parse_rdata {			## parse RR attributes in argument list
 	my $self = shift;
-	die join ' ', $self->type, 'not implemented' if ref($self) eq __PACKAGE__;
+	die join ' ', 'type', $self->type, 'not implemented' if ref($self) eq __PACKAGE__;
 	die join ' ', 'no zone file representation defined for', $self->type;
 }
 
@@ -548,13 +544,13 @@ Resource record data section when viewed as opaque octets.
 sub rdata {
 	my $self = shift;
 
-	return eval { $self->_encode_rdata( 0x4000, {} ); } unless scalar @_;
+	return eval { $self->_empty ? '' : $self->_encode_rdata( 0x4000, {} ); } || '' unless @_;
 
-	my $rdata = shift;
+	my $rdata = shift || '';
 	my $rdlen = $self->{rdlength} = length $rdata;
 	my $hash  = {};
 	$self->_decode_rdata( \$rdata, 0, $hash ) if $rdlen;
-	croak 'found compression pointer in rdata' if keys %$hash;
+	croak 'unexpected compression pointer in rdata' if keys %$hash;
 }
 
 
@@ -569,7 +565,7 @@ Returns a string representation of the RR-specific data.
 sub rdstring {
 	my $self = shift;
 
-	my @rdata = eval { $self->_format_rdata; };
+	my @rdata = eval { $self->_empty ? () : $self->_format_rdata; };
 	carp $@ if $@;
 
 	join "\n\t", _wrap(@rdata);
@@ -580,12 +576,12 @@ sub rdstring {
 
     $rdlength = $rr->rdlength;
 
-Returns the length of the encoded RR-specific data.
+Returns the uncompressed length of the encoded RR-specific data.
 
 =cut
 
 sub rdlength {
-	length shift->_encode_rdata;
+	length shift->rdata;
 }
 
 
@@ -673,19 +669,18 @@ our %_MINIMAL = ( 'ANY' => bless ['type' => 255], __PACKAGE__ );
 our %_LOADED = %_MINIMAL;
 
 sub _subclass {
-	my $class   = shift;
-	my $rrname  = shift;
-	my $default = shift;
+	my ( $class, $rrname, $default ) = @_;
 
 	unless ( $_LOADED{$rrname} ) {
-		local @INC = LIB;
 		my $rrtype = typebyname($rrname);
 
 		unless ( $_LOADED{$rrtype} ) {			# load once only
-			my $mnemon = typebyval($rrtype);
-			$mnemon =~ s/[^A-Za-z0-9]//g;		# expect the unexpected
+			local @INC = LIB;
 
-			my $subclass = join '::', __PACKAGE__, $mnemon;
+			my $identifier = typebyval($rrtype);
+			$identifier =~ s/\W/_/g;		# kosher Perl identifier
+
+			my $subclass = join '::', __PACKAGE__, $identifier;
 
 			unless ( eval "require $subclass" ) {
 				push @INC, sub {
@@ -723,6 +718,13 @@ sub _annotation {
 }
 
 
+my %ignore = map( ( $_ => 1 ), @core, 'annotation', '#' );
+
+sub _empty {
+	( $_[0]->{'#'} ||= scalar grep !$ignore{$_}, keys %{$_[0]} ) == 0;
+}
+
+
 sub _wrap {
 	my @text = @_;
 	my $cols = 80;
@@ -760,20 +762,19 @@ sub AUTOLOAD {				## Default method
 
 	my $string = $self->string;
 	my @object = grep defined($_), $oref, $oref->VERSION;
-	my $module = join '::', __PACKAGE__, typebyval( $self->{type} );
+	my $module = join '::', __PACKAGE__, $self->type;
 	eval("require $module") if $oref eq __PACKAGE__;
 
-	@_ = (<<"END");
-***  FATAL PROGRAM ERROR!!	Unknown method '$method'
+	@_ = ( <<"END", $@, "@object" );
+***  FATAL PROGRAM ERROR!!	Unknown instance method '$method'
 ***  which the program has attempted to call for the object:
 ***
 $string
 ***
-***  @object has no instance method '$method'
-***  $@
 ***  THIS IS A BUG IN THE CALLING SOFTWARE, which incorrectly assumes
 ***  that the object would be of a particular type.  The type of an
 ***  object should be checked before calling any of its methods.
+***
 END
 	goto &{'Carp::confess'};
 }
